(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2017 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`DensityMatrix`", {
    "AtomicDensityMatrix`",
    "AtomicDensityMatrix`Common`",
    "AtomicDensityMatrix`AtomicSystem`",
    "AtomicDensityMatrix`ExperimentalSystem`"
}];

Begin["`Private`"];

Options[Ket] = {TimeDependence -> Automatic, TimeVariable -> Automatic, ComplexExpandVariables -> Automatic};

Ket[syst : {_AtomicState..}, symbol_, opts : OptionsPattern[]] := Module[{ket, newopts, cev, td, tv},
    newopts = removeAutomaticOptions[Ket, opts];
    {cev, td, tv} = OptionValue[DensityMatrix, newopts, {ComplexExpandVariables, TimeDependence, TimeVariable}];
    ket = Switch[cev,
        True,
        Re@# + I Im@#& /@ (Subscript[symbol, SublevelLabel@#]& /@ syst),
        Subscript,
        Subscript[symbol, Re, SublevelLabel@#] + I Subscript[symbol, Im, SublevelLabel@#]& /@ syst,
        _,
        Subscript[symbol, SublevelLabel@#]& /@ syst
    ];
    If[td, ket = ket /. x : Subscript[symbol, __] -> x[tv]];
    Transpose[{ket}]
];

Options[DensityMatrix] = {
    TimeDependence -> True,
    Representation -> "Zeeman",
    DMSymbol -> \[Rho],
    DMLabel -> None,
    ComplexExpandVariables -> False,
    TimeVariable -> t
};

DMElement[sym_, part_, lab_, s1_, s2_] :=
    Subscript[sym, part /. {} -> Sequence[], lab /. (None | False -> Sequence[]), s1, s2];

DMElement[sym_, part_, lab_, s1_, s2_, time_] :=
    DMElement[sym, part, lab, s1, s2][time];

DMElement[sym_, part_, lab_, s1_, s2_, time_, d_] :=
    Derivative[d][DMElement[sym, part, lab, s1, s2]][time];

ExpandDMVariablesRule[syst_, opts : OptionsPattern[]] :=
    Module[
        {sll, patt, rep, imp, elem, rept, impt, rep1, imp1, rept1, impt1, elemt, e, r, i, r1, i1, cev, sym, lab},
        {cev, sym, lab} = OptionValue[DensityMatrix, {opts}, {ComplexExpandVariables, DMSymbol, DMLabel}];
        sll = SublevelLabel /@ syst;
        patt = DMElementPattern[ComplexExpandVariables -> False, TimeDependence -> All, FilterRules[{opts}, Options@DMElementPattern]];
        Switch[cev,
            True,
            (v : patt :> Switch[
                e = If[{Time} == {}, elem, elemt];
                Order[Position[sll, State1, {1}], Position[sll, State2, {1}]],
                -1, Re[e] - I Im[e],
                0, Re[v],
                1, Re[v] + I Im[v]
            ]) /. {
                elem -> DMElement[sym, {}, lab, State2, State1],
                elemt -> DMElement[sym, {}, lab, State2, State1, Time, TimeDerivative]
            },
            Subscript,
            (v : patt :> Switch[
                {r, i, r1, i1} = If[{Time} == {}, {rep, imp, rep1, imp1}, {rept, impt, rept1, impt1}];
                Order[Position[sll, State1, {1}], Position[sll, State2, {1}]],
                -1, r1 - I i1,
                0, r,
                1, r + I i
            ]) /. {
                rep -> DMElement[sym, Re, lab, State1, State2],
                imp -> DMElement[sym, Im, lab, State1, State2],
                rept -> DMElement[sym, Re, lab, State1, State2, Time, TimeDerivative],
                impt -> DMElement[sym, Im, lab, State1, State2, Time, TimeDerivative],
                rep1 -> DMElement[sym, Re, lab, State2, State1],
                imp1 -> DMElement[sym, Im, lab, State2, State1],
                rept1 -> DMElement[sym, Re, lab, State2, State1, Time, TimeDerivative],
                impt1 -> DMElement[sym, Im, lab, State2, State1, Time, TimeDerivative]
            },
            _,
            {}
        ]
    ];

DensityMatrix[syst : {_AtomicState..}, opts : OptionsPattern[]] :=
    Module[{rfunc},
        Switch[OptionValue@Representation,
            "Zeeman",
            rfunc[l1_, l2_] = If[OptionValue@TimeDependence,
                DMElement[OptionValue@DMSymbol, {}, OptionValue@DMLabel, l1, l2, OptionValue@TimeVariable],
                DMElement[OptionValue@DMSymbol, {}, OptionValue@DMLabel, l1, l2]
            ];
            DirectSelfProduct[rfunc, SublevelLabel /@ syst] /. ExpandDMVariablesRule[syst, opts],
            PolarizationMoments | "PolarizationMoments",
            PolarizationMoments[syst, opts]
        ]
    ];

DensityMatrix[as : {_AtomicState..}, es : {_ExperimentalRegion..}, opts : OptionsPattern[]] :=
    DensityMatrix[as, DMLabel -> RegionLabel[#], opts]& /@ es;

Options[DMVariables] = {
    TimeDependence -> Automatic,
    Representation -> Automatic,
    DMSymbol -> Automatic,
    DMLabel -> Automatic,
    ComplexExpandVariables -> Automatic,
    TimeVariable -> Automatic};

DMVariables[syst : {_AtomicState..}, opts : OptionsPattern[]] :=
    Module[{cev, newopts},
        newopts = removeAutomaticOptions[DMVariables, opts];
        cev = OptionValue[DensityMatrix, newopts, ComplexExpandVariables];
        If[cev || cev === Subscript,
            DeleteCases[Flatten@ERI@UpperTriangle[#], 0]&,
            Flatten
        ]@DensityMatrix[syst, newopts]
    ];

DMVariables[as : {_AtomicState..}, es : {_ExperimentalRegion..}, opts___Rule] :=
    Flatten[DMVariables[as, DMLabel -> RegionLabel[#], opts]& /@ es];

Quiet[LevelLabelPattern[State_, StateLabel_, F_, M_] :=
    State : {StateLabel_, F_, M_} | {StateLabel_, M_} | StateLabel_?(Head[#] =!= List&), {RuleDelayed::"rhs"}];

LevelLabelPattern1[] := LevelLabelPattern[State1, StateLabel1, F1, M1];

LevelLabelPattern2[] := LevelLabelPattern[State2, StateLabel2, F2, M2];

Options[DMElementPattern] = {DMSymbol -> Automatic, TimeDependence -> Automatic, ComplexExpandVariables -> Automatic};

DMElementPattern[opts : OptionsPattern[]] := Module[{newopts, r},
    newopts = removeAutomaticOptions[DMElementPattern, opts];
    r = OptionValue[DensityMatrix, newopts, DMSymbol];
    Switch[OptionValue[DensityMatrix, newopts, {ComplexExpandVariables, TimeDependence}],
        {Subscript, False}, dmElementPartPattern[r],
        {Subscript, True}, dmElementPartTDPattern[r],
        {Subscript, All}, dmElementPartPattern[r] | dmElementPartTDPattern[r],
        {_, False}, dmElementPattern[r],
        {_, True}, dmElementTDPattern[r],
        {_, All}, dmElementPattern[r] | dmElementTDPattern[r]
    ]
];

dmElementPattern[r_] = (
    Subscript[r, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]] |
        Subscript[r, LevelLabelPattern1[], LevelLabelPattern2[]]
) /; (DMLabel =!= Re && DMLabel =!= Im);

dmElementTDPattern[r_] = (
    Derivative[TimeDerivative_][Subscript[r, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]]][Time_] |
        Subscript[r, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]][Time_] |
        Derivative[TimeDerivative_][Subscript[r, LevelLabelPattern1[], LevelLabelPattern2[]]][Time_] |
        Subscript[r, LevelLabelPattern1[], LevelLabelPattern2[]][Time_]
) /; (DMLabel =!= Re && DMLabel =!= Im);

dmElementPartPattern[r_] = (
    Subscript[r, ReImPart : Re | Im, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]] |
        Subscript[r, ReImPart : Re | Im, LevelLabelPattern1[], LevelLabelPattern2[]]
) /; (DMLabel =!= Re && DMLabel =!= Im);

dmElementPartTDPattern[r_] = (
    Derivative[TimeDerivative_][Subscript[r, ReImPart : Re | Im, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]]][Time_] |
        Subscript[r, ReImPart : Re | Im, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]][Time_] |
        Derivative[TimeDerivative_][Subscript[r, ReImPart : Re | Im, LevelLabelPattern1[], LevelLabelPattern2[]]][Time_] |
        Subscript[r, ReImPart : Re | Im, LevelLabelPattern1[], LevelLabelPattern2[]][Time_]
) /; (DMLabel =!= Re && DMLabel =!= Im);

ExprToReIm[expr_List, opts : OptionsPattern[]] := Flatten[ExprToReIm[#, opts]& /@ expr];

ExprToReIm[expr_, opts : OptionsPattern[]] :=
    DeleteCases[
        Chop@Refine[
            ComplexExpand[
                Flatten@Through[{Re, Im}@expr] /. {(f : Re | Im)[x_ == y_] :> f[x] == f[y]},
                DMElementPattern[ComplexExpandVariables -> False, opts, TimeDependence -> All]
            ],
            DMElementPattern[ComplexExpandVariables -> Subscript, opts, TimeDependence -> All] \[Element] Reals
        ] /. {
            Re[x : DMElementPattern[ComplexExpandVariables -> Subscript, opts, TimeDependence -> All]] -> x,
            Im[x : DMElementPattern[ComplexExpandVariables -> Subscript, opts, TimeDependence -> All]] -> 0
        },
        True | 0
    ];

SetAttributes[ERI, Listable];

ERI[x_ == y_] := MapThread[Equal, ERI@{x, y}];

ERI[x_] :=
    Total@Flatten@# & /@ Reap[ERISow[Expand[x /. y : Exp[_] :> ExpToTrig[y]]], {Re, Im}][[2]];

ERISow[x_Plus] := Scan[ERISow, x];

ERISow[x_] :=
    If[
        FreeQ[x, Complex],
        Sow[x, Re],
        If[
            MatchQ[x, Complex[0, a_] b_. /; FreeQ[b, Complex]],
            Sow[x / I, Im],
            Sow[ComplexExpand[Re[x], TargetFunctions -> {Re, Im}], Re];
            Sow[ComplexExpand[Im[x], TargetFunctions -> {Re, Im}], Im];
        ]
    ];

End[];

EndPackage[];
