(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2017 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`LiouvilleEquation`", {
(*"PhysicalConstants`","Units`",*)
    "AtomicDensityMatrix`",
    "AtomicDensityMatrix`Common`",
    "AtomicDensityMatrix`CouplingCoefficients`",
    "AtomicDensityMatrix`TensorAlgebra`",
    "AtomicDensityMatrix`AtomicSystem`",
    "AtomicDensityMatrix`ExperimentalSystem`",
    "AtomicDensityMatrix`DensityMatrix`",
    "AtomicDensityMatrix`WignerEckart`",
    "AtomicDensityMatrix`WignerRotate`"}];

Begin["`Private`"];

LiouvilleRHS[system : {_AtomicState..}, H_?OperatorQ, GMatrix_?OperatorQ, LMatrix_?OperatorQ, opts : OptionsPattern[]] :=
    -I Commutator[H, DensityMatrix[system, opts]] - Anticommutator[DensityMatrix[system, opts], GMatrix] / 2 + LMatrix;

Options[LiouvilleEquation] = Options[DMVariables];

LiouvilleEquation[system : {_AtomicState..}, H_?OperatorQ, GMatrix_?OperatorQ, LMatrix_?OperatorQ, opts : OptionsPattern[]] :=
    Module[{rep, lhs, rhs, newopts, tv, cev, H1, GMatrix1, LMatrix1},
        newopts = removeAutomaticOptions[LiouvilleEquation, opts];
        {rep, tv, cev} = OptionValue[DensityMatrix, newopts, {Representation, TimeVariable, ComplexExpandVariables}];
        lhs = D[DensityMatrix[system, newopts], tv];
        {H1, GMatrix1, LMatrix1} = {H, GMatrix, LMatrix} /. ExpandDMVariablesRule[system, newopts];
        rhs = LiouvilleRHS[system, H1, GMatrix1, LMatrix1, newopts];
        If[rep === PolarizationMoments,
            MapThread[Equal, {Simplify@Flatten@Decomposition[system, lhs], Flatten@Decomposition[system, rhs]}, 1],
            If[cev || cev === Subscript,
                DeleteCases[Flatten@ERI@UpperTriangle[#], True]&,
                Flatten
            ]@MapThread[Equal, {lhs, rhs}, 2]
        ]
    ];

LiouvilleEquation[system : {_AtomicState..}, H_?OperatorQ, GMatrix_?OperatorQ, opts : OptionsPattern[]] :=
    LiouvilleEquation[system, H, GMatrix, 0 IdentityMatrix@Length@system, opts];

LiouvilleEquation[system : {_AtomicState..}, H_?OperatorQ, opts : OptionsPattern[]] :=
    LiouvilleEquation[system, H, 0 IdentityMatrix@Length@system, 0 IdentityMatrix@Length@system, opts];

LiouvilleEquation[
    as : {_AtomicState..},
    es : {_ExperimentalRegion..},
    H : {_?OperatorQ..},
    GMatrix : {_?OperatorQ..},
    LMatrix : {_?OperatorQ..},
    opts : OptionsPattern[]
] :=
    Flatten[
        MapThread[LiouvilleEquation[as, #2, #3, #4, DMLabel -> RegionLabel[#1], opts]&, {es, H, GMatrix, LMatrix}],
        1
    ];

LiouvilleEquation[sys : {_AtomicState..}, H_?ContravariantDecompositionMatrixQ,
    G_?ContravariantDecompositionMatrixQ, L_?ContravariantDecompositionMatrixQ, opts : OptionsPattern[]
] :=
    Module[{
        newopts = removeAutomaticOptions[LiouvilleEquation, opts],
        tv,
        levels = RemoveZeemanSublevels[sys], pm, n, f, component
    },
        tv = OptionValue[DensityMatrix, newopts, TimeVariable];
        pm = PolarizationMoments[levels, opts];
        n = Length[levels];
        f = F /@ levels;
        component[op_, {i1_, i2_}, k_, q_] :=
            op[[i1, i2, k - Abs[f[[i1]] - f[[i2]]] + 1, -q + k + 1, 1]];
        Flatten@Table[
            Transpose@{Table[
                D[component[pm, {i1, i2}, k, q], tv] ==
                    Sum[
                        -I (-1)^(f[[i1]] + f[[i2]] + k) Sqrt[(2k1 + 1)(2k2 + 1)]CG[{k1, q1}, {k2, q2}, {k, q}] *
                            SJS[{k1, k2, k}, {f[[i2]], f[[i1]], f[[i3]]}] *
                                ((component[H, {i1, i3}, k1, q1] - I / 2 component[G, {i1, i3}, k1, q1])component[pm, {i3, i2}, k2, q2]
                                    - component[pm, {i1, i3}, k1, q1](component[H, {i3, i2}, k2, q2] + I / 2 component[G, {i3, i2}, k2, q2])),
                        {i3, n},
                        {k1, Abs[f[[i1]] - f[[i3]]], f[[i1]] + f[[i3]]},
                        {q1, k1, -k1, -1},
                        {k2, Abs[f[[i2]] - f[[i3]]], f[[i2]] + f[[i3]]},
                        {q2, k2, -k2, -1}
                    ] + component[L, {i1, i2}, k, q],
                {q, k, -k, -1}
            ]},
            {i1, n},
            {i2, n},
            {k, Abs[f[[i1]] - f[[i2]]], f[[i1]] + f[[i2]]}
        ]
    ];

Options[InitialConditions] = DeleteCases[Options[DMVariables], TimeVariable -> _];

InitialConditions[system : {_AtomicState..}, InitialDM_?OperatorQ, t0_, opts : OptionsPattern[]]:=
    Module[{lhs, newopts, cev},
        newopts = removeAutomaticOptions[LiouvilleEquation, opts];
        cev = OptionValue[DensityMatrix, newopts, ComplexExpandVariables];
        lhs = DensityMatrix[system, TimeVariable -> t0, newopts];
        If[
            cev || cev === Subscript,
            ExprToReIm[UpperTriangle[#], newopts]&,
            Flatten
        ]@MapThread[Equal, {lhs, InitialDM}, 2]
    ];

InitialConditions[as : {_AtomicState..}, es : {_ExperimentalRegion..}, InitialDM : {_?OperatorQ..}, t0_, opts : OptionsPattern[]] :=
    Flatten[MapThread[InitialConditions[as, #2, t0, DMLabel -> RegionLabel[#1], opts]&, {es, InitialDM}], 1];

End[];

EndPackage[];
